<?php
// $Id: common.mapclient.inc,v 1.1.2.3 2009/01/27 21:18:37 openwereld Exp $

function mapclient_replace_map_tags($matches) {
  global $logger;
  
  $logger->debug('replace ', $matches);
  global $map_count;
  if (! isset($map_count)) {
    $map_count = 0;
  }
  
  // Correct for encoding " as &quot; by fckeditor/tinymce
  for ($i = 1, $len = count($matches); $i < $len; $i++) {
    $matches[$i] = html_entity_decode($matches[$i]);
  }
  
  $attrs = _mapclient_get_map_attrs($matches);
  $inline_map_spec = $matches[0];
  $logger->debug('map attrs ', $attrs);
  
  if (!isset($attrs['id'])) {
    $attrs['id'] = 'map-'. $map_count;
  }
  
  // add mapclient class
  //$attrs['class'] = (empty($attrs['class']) ? '' : $attrs['class'] . ' ') . 'mapclient';
  
  // Use internal mapclient for print pages. (also book/export?)
  if (strncmp(arg(0), 'print', 5) == 0 && !isset($attrs['mapclient'])) {
    $attrs['mapclient'] = 'internal';
  }
  
  // invoke init_header with (almost) unmodified attrs
  mapclient_invoke('init_header', $attrs);
  
  // Determine maxextent/extent/zoom/center/projection of map from baselayers, contexts, 
  // layers and explicit settings as attributes
  $options = array();
  
  // (1) baselayers
  // If baselayers contain OpenStreetMap the projection, maxresolution and extent are fixed
  $baselayers_texts = array();
  if (isset($attrs['baselayers'])) {
    $baselayers_texts[] = $attrs['baselayers'];
  }
  
  // (2) contexts
  if (isset($attrs['context']) && !isset($attrs['contexts'])) {
    $attrs['contexts'] = $attrs['context'];
    unset($attrs['context']);
  }
  if (isset($attrs['contexts'])) {
    $contexts = explode(',', $attrs['contexts']);
    // check the first context
    if (1 == preg_match('/node\/(?P<nid>\d+)(\/revisions\/(?P<vid>\d+))?/', $contexts[0], $matches)) {
      // get extent/projection from context
      $node = node_load($matches['nid']
        , isset($matches['vid']) ? $matches['vid'] : NULL);
      if (!$node) {
        drupal_set_message(t('Invalid node in map: %nid.  Please correct the inline map: %inline_map.', array('%nid' => $matches['nid'], '%inline_map' => $inline_map_spec)), 'error');
      }
      else if ($node->type == 'mapcontext') {
        //drupal_set_message(var_export($node, true));
        // a valid mapcontext node
        if (!isset($options['width'])) {
          $options['width'] = $node->width;
        }
        if (!isset($options['height'])) {
          $options['height'] = $node->height;
        }
        
        $srs = 'EPSG:'. $node->srid;
        if (!isset($options['projection'])) {
          $options['projection'] = $srs;
        }
        
        $maxextent = $node->minx .','. $node->miny .','. $node->maxx .','. $node->maxy;
        if (!empty($node->extension['baselayers'])) {
          $baselayers_texts[] = $node->extension['baselayers'];
          
          if ($maxextent != '-180,-90,180,90'
          && $maxextent != '-20037508.34,-20037508.34,20037508.34,20037508.34') {
            // use extent as restricted extent
            $options['restrictedextent'] = $maxextent;
          }
        }
        else {
          // no baselayers so use extent as maxextent
          if (!isset($options['maxextent'])) {
            $options['maxextent'] = $maxextent;
          }
        }
        if (!isset($options['extent']) && is_numeric($node->dminx)) {
          $options['extent'] = $node->dminx .','. $node->dminy .','. $node->dmaxx .','. $node->dmaxy;
        }
      }
      else {
        drupal_set_message(t('Node %nid is not a map context. Please correct the inline map: %inline_map.', array('%nid' => $node->nid, '%inline_map' => $inline_map_spec)), 'error');
      }
    }
  }
  
  if (count($baselayers_texts) > 0) {
    $baselayers_text = implode(',', $baselayers_texts);
    $logger->info('baselayers text: '. $baselayers_text);
    $baselayers = get_baselayers($baselayers_text);
    foreach ($baselayers as $baselayer) {
      switch ($baselayer['baselayer_vendor']) {
      case 'oam':
      case 'osm':
        // openaerialmap and openstreetmap always use spherical mercator
        $logger->info('osm/oam => internalprojection = EPSG:900913');
        $attrs['internalprojection'] = 'EPSG:900913';
        break;
      }
    }
  }
  
  // (3) layers
  // Map layers can be defined as a list of layers with each layer
  // defined as type|url|etc.
  // E.g. georss|url|title|marker or gpx|url|title or
  // wms|url|title|names.
  if (isset($attrs['layers'])) {
    $layers = explode(',', $attrs['layers']);
    if (count($layers) > 0 && count($baselayers_texts) > 0 && $attrs['baselayers'] != 'none') {
      $logger->info('baselayers with overlay => internalprojection = EPSG:900913');
      $options['internalprojection'] = 'EPSG:900913';
      //$options['maxextent'] = '-20037508.34, -20037508.34, 20037508.34, 20037508.34';
      //$options['sphericalMercator'] = 'true';
    }
    foreach ($layers as $lid) {
      if (!isset($attrs[$lid])) {
        drupal_set_message(t("Layer definition '%lid' is missing", array('%lid' => $lid)), 'error');
        continue;
      }
      /*
      $layer = explode('|', $attrs[$lid]);
      switch ($layer[0]) {
      case 'gbkn':
        break;
      case 'georss':
        break;
      case 'gml':
        break;
      case 'gpx':
        break;
      case 'wfs':
        break;
      case 'wms':
        break;
      }*/
    }
  }
  
  if (!isset($options['width'])) {
    $options['width'] = variable_get('mappingkit_map_width', DEFAULT_MAPPINGKIT_MAP_WIDTH); // default width
  }
  if (!isset($options['height'])) {
    $options['height'] = variable_get('mappingkit_map_height', DEFAULT_MAPPINGKIT_MAP_HEIGHT); // default height
  }
  
  // copy options into attrs
  foreach ($options as $key => $value) {
    if (!isset($attrs[$key])) {
      $attrs[$key] = $value;
    }
  }
  
  $attrs['defaults'] = $options;
  $logger->debug('attrs', $attrs);
  
  // standard style
  $attrs['style'] = (empty($attrs['style']) ? '' : $attrs['style'] . '; ') . "width: {$attrs['width']}px; height: {$attrs['height']}px;";
  
  // render body html of map
  $output = mapclient_invoke('render_body', $attrs);
  
  global $maps;
  $maps[ $attrs['id'] ] = $attrs;
  $map_count += 1;
  
  return $output;
}

/**
 * Invoke functions for the mapclient of the current user or the
 * default mapclient.
 */
function mapclient_invoke($hook, &$args) {
  static $finished = array();
  static $initialized = array();
  global $user;
  global $logger;
  
  if (isset($_GET['mapclient'])) {
    // use mapclient given in url parameters
    $mapclient = $_GET['mapclient'];
    $logger->info("mapclient {$args['id']} override by url parameter");
  }
  else if (isset($args['mapclient'])) {
    // use mapclient set in attribute
    $mapclient = $args['mapclient'];
    $logger->info("mapclient {$args['id']} override by map attribute");
  }
  else if (isset($user->mapclient)) {
    $mapclient = $user->mapclient;
    $logger->info("mapclient {$args['id']} override by user");
  }
  else {
    // default mapclient for user or system
    $mapclient = variable_get('mapclient', DEFAULT_MAPCLIENT);
  }
  
  // get specific mapclient info to determine if a file needs
  // to be included
  $mapclients = mapclient_get_clients();
  $info = $mapclients[$mapclient];
  if (isset($info['include name'])) {
    // include mapclient code
    module_load_include('inc', isset($info['module']) ? $info['module'] : $mapclient
      , $info['include name']);
    
    // include mapclient javascript code (must be included after mapclient.js)
    $scope = variable_get('mappingkit_add_js_scope', DEFAULT_MAPPINGKIT_ADD_JS_SCOPE);
    drupal_add_js(drupal_get_path('module', $info['module']) . '/' . $info['include name'] . '.js'
      , 'extra', $scope);
  }
  
  if ($hook == 'init_header') {
    // set/overwrite mapclient
    $args['mapclient'] = $mapclient;

    if (isset($initialized[$mapclient])) {
      // already initialized
      $logger->info("$mapclient already initialized");
      return;
    }
    $initialized[$mapclient] = TRUE;
  }
  else if ($hook == 'finish') {
    if (isset($finished[$mapclient])) {
      // already finished
      $logger->info("$mapclient already finished");
      return;
    }
    $finished[$mapclient] = TRUE;
  }
  
  // invoke function
  $function = (isset($info['module']) ? $info['module'] : $mapclient) .'_'. $hook;
  if (function_exists($function)) {
    return $function($args);
  }
  else {
    drupal_set_message("$function does not exist", "error");
  }
}

function _mapclient_get_map_attrs($matches) {
  global $logger;
  
  $element_value = '';
  if( isset($matches[MATCH_OPEN_MAP_VALUE]) && $matches[MATCH_OPEN_MAP_VALUE] != '' ) {
    // format <map>relative/path/to/config</map>
    $element_value = $matches[MATCH_OPEN_MAP_VALUE];
    
    $attrs = array();
    if( isset($matches[MATCH_OPEN_MAP_ATTRIBUTES]) && $matches[MATCH_OPEN_MAP_ATTRIBUTES] != '' ) {
      // format <map config="relative/path/to/config"></map>
      $attribs = $matches[MATCH_OPEN_MAP_ATTRIBUTES];
      $n = preg_match_all( variable_get('mapclient_attrib_expr',DEFAULT_MAPCLIENT_ATTRIB_EXPR), $attribs, $match_attrs );
      // combine separate key and value arrays to assoc array of key-value pairs
      $attrs = array_combine( $match_attrs[1], $match_attrs[2] );
    }
  }
  else {
    // check attributes
    if( isset($matches[MATCH_OPEN_MAP_ATTRIBUTES]) && $matches[MATCH_OPEN_MAP_ATTRIBUTES] != '' )
      // format <map config="relative/path/to/config"></map>
      $attribs = $matches[MATCH_OPEN_MAP_ATTRIBUTES];
    else
      // format <map config="relative/path/to/config"/>
      $attribs = $matches[MATCH_CLOSED_MAP_ATTRIBUTES];
    
    $n = preg_match_all( variable_get('mapclient_attrib_expr',DEFAULT_MAPCLIENT_ATTRIB_EXPR), $attribs, $match_attrs );
    // combine separate key and value arrays to assoc array of key-value pairs
    $attrs = array_combine($match_attrs[1], $match_attrs[2]);
  }
  // value of <map> element
  $attrs['value'] = $element_value;
  
  return $attrs;
}

/**
 * Explode options string into associative array of options.
 * @param $options_str
 *   String with option keys and values. Key=value pairs are separated by 
 *   comma's. Note that the value may contain '='-signs.
 *   Format example:
 *     key_a=value_a,key_b={key_c:value,key_d:value},key_c=x=1&y=2.
 * @return Associative array of options and their values.
 */
function mk_explode_options($options_str) {
  //drupal_set_message($options_str);
  $options = array();
  $kv_regex = '/(\w+)\=(\{.*\}|[^,\{\}]*?)/Us';
  $pv_regex = '/(\w+)\:([^,\}]*?)/Us';
  $result = preg_match_all($kv_regex, $options_str, $matches);
  //drupal_set_message('matches: '.var_export($matches, true));
  for ($i = 0; $i < $result; $i++) {
    $key = $matches[1][$i];
    $value = $matches[2][$i];
    // check if value contains property:value pairs
    if ($value[0] == '{' && ($pcount = preg_match_all($pv_regex, $value, $pmatches)) > 0) {
      $options[$key] = array_combine($pmatches[1], $pmatches[2]);
    }
    else {
      $options[$key] = $value;
    }
  }
  //drupal_set_message('options: ' . var_export($options, true));
  return $options;
}

/**
 *
 */
function get_baselayers($baselayers_str) {
  global $baselayers_separator, $baselayer_options_separator;
  
  $baselayers = array();
  $baselayers_arr = explode($baselayers_separator, $baselayers_str);
  $i = 0;
  foreach($baselayers_arr as $baselayer_info) {
    $tmp = explode($baselayer_options_separator, $baselayer_info);
    $baselayer_variant = $tmp[0];
    $baselayer_options = (count($tmp) > 1) ? array_splice($tmp, 1) : array();
    $tmp = explode('-', $baselayer_variant);
    $baselayer_vendor = $tmp[0];
    $variant = (count($tmp) > 1) ? strtolower($tmp[1]) : 'default';
    $baselayers[$i]['baselayer_vendor'] = $baselayer_vendor;
    $baselayers[$i]['variant'] = $variant;
    $baselayers[$i]['baselayer_options'] = get_baselayer_options($baselayer_options);
    $i++;
  }
  return $baselayers;
}

/**
 *
 */
function get_baselayer_options($options) {
  $baselayer_options = array();
  foreach($options as $option) {
    $tmp = explode('=', $option);
    $baselayer_options[$tmp[0]] = $tmp[1];
  }
  return $baselayer_options;
}

function mapclient_common_open_outer_block($attrs) {
  $output = '<div id="map-'. $attrs['id'] .'-block" class="map-block ' . $attrs['class'] . '">';
  return $output;
}

function mapclient_common_close_outer_block() {
  return '</div>';
}

function mapclient_common_open_map_block($attrs, $options = array()) {
  $style = "width: {$attrs['width']}px; height: {$attrs['height']}px;";
  if (!empty($attrs['style'])) {
    $style .= $attrs['style'];
  }
  $class_param = empty($options['class']) ? '' : 'class="' . $options['class'] . '" ';
  //$output = '<div id="'. $attrs['id'] .'" class="mapclient" style="'. $style .'">';
  $output = '<div id="'. $attrs['id'] .'" ' . $class_param . 'style="'. $style .'">';
  return $output;
}

function mapclient_common_close_map_block() {
  return '</div>';
}

function mapclient_common_render_controls($attrs) {
  
  $clients = mapclient_get_clients();
  $info = $clients[$attrs['mapclient']];
  
  $output = '';
  
  if (isset($attrs['zoomlist'])) {
    //$output .= '<select onchange="maps[0].zoomToExtent(OpenLayers.Bounds.fromString(this.value));">';
    $output .= "<select onchange=\"MappingKit.MapClient.{$info['name']}.zoomToExtent('{$attrs['id']}', this.value);\">"; //"
    $zoomlist = mk_explode_options($attrs['zoomlist']);
    //drupal_set_message(var_export($zoomlist, true));
    
    // TODO: count($zoomlist) > 1 -> optgroup for queries
    
    foreach ($zoomlist as $key => $value) {
      if (0 != preg_match('/\{(?P<extent>(?P<minx>[\d\.\-]+),(?P<miny>[\d\.\-]+),(?P<maxx>[\d\.\-]+),(?P<maxy>[\d\.\-]+))(,(?P<text>.*))?\}/Us', $value, $matches)) { //"
        // single option
        //drupal_set_message(var_export($matches, true));
        $output .= '<option value="' . $matches['extent'] . '">';
        $output .= empty($matches['text']) ? $key : $matches['text'];
        $output .= '</option>';
      }
      else if (0 != preg_match('/\{((?P<database>\w+)\|)?(?P<query>select .*)\}/Us', $value, $matches)) {
        // query option
        //drupal_set_message(var_export($matches, true));
        if (!empty($matches['database'])) {
          $prev_db = db_set_active($matches['database']);
        }
        else if (module_exists('geodata')) {
          $prev_db = db_set_active($db = variable_get('geodata_database', DEFAULT_GEODATA_DATABASE));
        }
        
        $tmp = explode('|', $matches['query']);
        $matches['query'] = $tmp[0];
        if (count($tmp) > 1) {
          $matches['group'] = $tmp[1];
        }
        
        //drupal_set_message('query: ' . $matches['query']);
        $result = db_query($matches['query']);
        
        if (!empty($matches['group'])) {
          // Add option group
          $output .= '<optgroup label="' . $matches['group'] . '">';
        }
        
        $count = 0;
        while ($row = db_fetch_array($result)) {
          // Expect format (maxx,maxy),(minx,miny) as from st_box(geom)
          // or BOX(minx miny,maxx maxy) as from box2d(geom)
          if (0 != preg_match('/BOX\((?P<minx>[\d\.\-]+) (?P<miny>[\d\.\-]+),(?P<maxx>[\d\.\-]+) (?P<maxy>[\d\.\-]+)\)/', $row['extent'], $matches)
          || 0 != preg_match('/\((?P<maxx>[\d\.\-]+),(?P<maxy>[\d\.\-]+)\),\((?P<minx>[\d\.\-]+),(?P<miny>[\d\.\-]+)\)/', $row['extent'], $matches)) {
            //drupal_set_message(var_export($matches, true));
            $extent = $matches['minx'] . ',' . $matches['miny'] . ',' . $matches['maxx'] . ',' . $matches['maxy'];
          }
          else {
            // unknown data format so other records will probably not be valid?
            drupal_set_message(t('Unknown format for extent in query of zoomlist.'));
            break;
          }
          
          $output .= '<option value="' . $extent . '">';
          $output .= empty($row['label']) ? ($key . ' ' . ++$count) : $row['label'];
          $output .= '</option>';
        }
        if (!empty($matches['group'])) {
          $output .= '</optgroup>';
        }
        if ($prev_db) {
          db_set_active($prev_db);
        }
      }
    }
    $output .= "</select>";//"
  }
  return $output;
}

